// lex2_test.cpp
// Glenn G. Chappell
// 6 Feb 2013
//
// For CS 331 Spring 2013
// Test program for class Lex2
// Used in Assignment 2, Exercise A

// Includes for code to be tested
#include "lex2.h"       // For class Lex2
#include "lex2.h"       // Double inclusion test

// Includes for testing package & code common to all test programs
#include <iostream>     // for std::cout, std::endl, std::cin
#include <string>       // for std::string
#include <stdexcept>    // for std::runtime_error

// Additional includes for this test program
#include <vector>       // for std::vector
#include <cstdlib>      // for std::size_t
#include <stdexcept>    // for std::exception

// If the constant SINGLE_TEST is defined, then this program terminates
// after the first failed test (if any). Comment out the following line
// to have the program run all tests, regardless.
#define SINGLE_TEST


// ************************************************************************
// Testing Package:
//     Class Tester - For Tracking Tests
// ************************************************************************


// class Tester
// For extremely simple unit testing.
// Keeps track of number of tests and number of passes.
// Use test (with success/failure parameter) to do a test.
// Get results with numTests, numPassed, numFailed, allPassed.
// Restart testing with reset.
// Invariants:
//     countTests_ == number of tests (calls to test) since last reset.
//     countPasses_ == number of times function test called with true param
//      since last reset.
//     0 <= countPasses_ <= countTests_.
//     tolerance_ >= 0.
class Tester {

// ***** Tester: ctors, dctor, op= *****
public:

    // Default ctor
    // Sets countTests_, countPasses_ to zero, tolerance_ to given value
    // Pre: None.
    // Post:
    //     numTests == 0, countPasses == 0, tolerance_ == abs(theTolerance)
    // Does not throw (No-Throw Guarantee)
    Tester(double theTolerance = 0.0000001)
        :countTests_(0),
         countPasses_(0),
         tolerance_(theTolerance >= 0 ? theTolerance : -theTolerance)
    {}

    // Compiler-generated copy ctor, copy op=, dctor are used

// ***** Tester: general public functions *****
public:

    // test
    // Handles single test, param indicates pass/fail
    // Pre: None.
    // Post:
    //     countTests_ incremented
    //     countPasses_ incremented if (success)
    //     Message indicating test name (if given)
    //      and pass/fail printed to cout
    // Does not throw (No-Throw Guarantee)
    //  - Assuming exceptions have not been turned on for cout.
    void test(bool success,
              const std::string & testName = "")
    {
        ++countTests_;
        if (success) ++countPasses_;

        std::cout << "    ";
        if (testName != "")
        {
            std::cout << "Test: "
                      << testName
                      << " - ";
        }
        std::cout << (success ? "passed" : "********** FAILED **********")
                  << std::endl;
    }

    // ftest
    // Does single floating-point test.
    // Tests passes iff difference of first two values is <= tolerance.
    // Pre: None.
    // Post:
    //     countTests_ incremented
    //     countPasses_ incremented if (abs(val1-val2) <= tolerance_)
    //     Message indicating test name (if given)
    //      and pass/fail printed to cout
    // Does not throw (No-Throw Guarantee)
    void ftest(double val1,
               double val2,
               const std::string & testName = "")
    { test(val1-val2 <= tolerance_ && val2-val1 <= tolerance_, testName); }

    // reset
    // Resets *this to default constructed state
    // Pre: None.
    // Post:
    //     countTests_ == 0, countPasses_ == 0
    // Does not throw (No-Throw Guarantee)
    void reset()
    {
        countTests_ = 0;
        countPasses_ = 0;
    }

    // numTests
    // Returns the number of tests that have been done since last reset 
    // Pre: None.
    // Post:
    //     return == countTests_
    // Does not throw (No-Throw Guarantee)
    int numTests() const
    { return countTests_; }

    // numPassed
    // Returns the number of tests that have passed since last reset
    // Pre: None.
    // Post:
    //     return == countPasses_
    // Does not throw (No-Throw Guarantee)
    int numPassed() const
    { return countPasses_; }

    // numFailed
    // Returns the number of tests that have not passed since last reset
    // Pre: None.
    // Post:
    //     return + countPasses_ == numTests_
    // Does not throw (No-Throw Guarantee)
    int numFailed() const
    { return countTests_ - countPasses_; }

    // allPassed
    // Returns true if all tests since last reset have passed
    // Pre: None.
    // Post:
    //     return == (countPasses_ == countTests_)
    // Does not throw (No-Throw Guarantee)
    bool allPassed() const
    { return countPasses_ == countTests_; }

    // setTolerance
    // Sets tolerance_ to given value
    // Pre: None.
    // Post:
    //     tolerance_ = abs(theTolerance)
    // Does not throw (No-Throw Guarantee)
    void setTolerance(double theTolerance)
    { tolerance_ = (theTolerance >= 0 ? theTolerance : -theTolerance); }

// ***** Tester: data members *****
private:

    int countTests_;    // Number of tests done since last reset
    int countPasses_;   // Number of tests passed since last reset
    double tolerance_;  // Tolerance for floating-point near-equality tests

};  // end class Tester


// ************************************************************************
// Testing Package:
//     Class TypeCheck - Helper Class for Type Checking
// ************************************************************************


// class TypeCheck
// This class exists in order to have static member function check, which
// takes a parameter of a given type, by reference. Objects of type
// TypeCheck<T> cannot be created.
// Usage:
//     TypeCheck<MyType>::check(x)
//     returns true if the type of x is (MyType) or (const MyType),
//     otherwise false.
// Invariants: None.
// Requirements on Types: None.
template<typename T>
class TypeCheck {

private:

    // Uncopyable class. Do not define copy ctor, copy assn.
    TypeCheck(const TypeCheck &);
    TypeCheck<T> & operator=(const TypeCheck &);

    // Compiler-generated dctor is used (but irrelevant).

public:

    // check
    // The function and function template below simulate a single function
    // that takes a single parameter, and returns true iff the parameter has
    // type T or (const T).

    // check (reference-to-const T)
    // Pre: None.
    // Post:
    //     Return is true.
    // Does not throw (No-Throw Guarantee)
    static bool check(const T & param)
    { return true; }

    // check (reference-to-const non-T)
    // Pre: None.
    // Post:
    //     Return is false.
    // Requirements on types: None.
    // Does not throw (No-Throw Guarantee)
    template <typename OtherType>
    static bool check(const OtherType & param)
    { return false; }

};  // End class TypeCheck


// ************************************************************************
// Testing Package:
//     Class Counter - Helper Class for Counting Calls & Objects, Throwing
// ************************************************************************


// class Counter
// Item type for counting ctor, dctor, op= calls, counting existing
//  objects, and possibly throwing on copy. Has operator< (which always
//  returns false), allowing it to be the value type of a sorted container.
// If static member copyThrow_ is set, then copy ctor and copy assn throw
//  std::runtime_error. Exception object's "what" member is set to "C" by
//  the copy ctor and "A" by copy assn.
// Increments static data member ctorCount_ on default construction and
//  successful copy construction. Increments static data member assnCount_
//  on successful copy assignment. Increments static data member
//  dctorCount_ on destruction.
// Increments static data member existing_ on construction, and decrements
//  it on destruction.
// Static data member maxExisting_ is highest value of existing_ since last
//  reset, or start of program if reset has never been called.
// Invariants:
//     Counter::existing_ is number of existing objects of this class.
//     Counter::ctorCount_ is number of successful ctor calls since
//      most recent call to reset, or start of program if reset has never
//      been called.
//     Counter::dctorCount_ is (similarly) number of dctor calls.
//     Counter::assnCount_ is (similarly) number of copy assn calls.
//     Counter::maxExisting_ is (similarly) highest value existing_ has
//      assumed.
class Counter {

// ***** Counter: Ctors, dctor, op= *****
public:

    // Default ctor
    // Pre: None.
    // Post:
    //     (ctorCount_ has been incremented.)
    //     (existing_ has been incremented.)
    // Does not throw (No-Throw Guarantee)
    Counter()
    {
        ++existing_;
        if (existing_ > maxExisting_)
            maxExisting_ = existing_;
        ++ctorCount_;
    }

    // Copy ctor
    // Throws std::runtime_error if copyThrow_.
    // Pre: None.
    // Post:
    //     (ctorCount_ has been incremented.)
    //     (existing_ has been incremented.)
    // May throw std::runtime_error
    // Strong Guarantee
    Counter(const Counter & other)
    {
        if (copyThrow_)
            throw std::runtime_error("C");
        ++existing_;
        if (existing_ > maxExisting_)
            maxExisting_ = existing_;
        ++ctorCount_;
    }

    // Copy assignment
    // Throws std::runtime_error if copyThrow_.
    // Pre: None.
    // Post:
    //     Return value is *this.
    //     (assnCount_ has been incremented.)
    // May throw std::runtime_error
    // Strong Guarantee
    Counter & operator=(const Counter & rhs)
    {
        if (copyThrow_)
            throw std::runtime_error("A");
        ++assnCount_;
        return *this;
    }

    // Dctor
    // Pre: None.
    // Post:
    //     (dctorCount_ has been incremented.)
    //     (existing_ has been decremented.)
    // Does not throw (No-Throw Guarantee)
    ~Counter()
    {
        --existing_;
        ++dctorCount_;
    }

// ***** Counter: Functions dealing with count *****
public:

    // reset
    // Pre: None.
    // Post:
    //     maxExisting_ == existing_.
    //     ctorCount_ == 0.
    //     dctorCount_ == 0.
    //     assnCount_ == 0.
    //     copyThrow_ == shouldThrow.
    // Does not throw (No-Throw Guarantee)
    static void reset(bool shouldThrow = false)
    {
        maxExisting_ = existing_;
        ctorCount_ = 0;
        dctorCount_ = 0;
        assnCount_ = 0;
        copyThrow_ = shouldThrow;
    }

    // getExisting
    // Pre: None.
    // Post:
    //     return == existing_.
    // Does not throw (No-Throw Guarantee)
    static int getExisting()
    { return existing_; }

    // getMaxExisting
    // Pre: None.
    // Post:
    //     return == maxExisting_.
    // Does not throw (No-Throw Guarantee)
    static int getMaxExisting()
    { return maxExisting_; }

    // getCtorCount
    // Pre: None.
    // Post:
    //     return == ctorCount_.
    // Does not throw (No-Throw Guarantee)
    static int getCtorCount()
    { return ctorCount_; }

    // getDctorCount
    // Pre: None.
    // Post:
    //     return == dctorCount_.
    // Does not throw (No-Throw Guarantee)
    static int getDctorCount()
    { return dctorCount_; }

    // getAssnCount
    // Pre: None.
    // Post:
    //     return == assnCount_.
    // Does not throw (No-Throw Guarantee)
    static int getAssnCount()
    { return assnCount_; }

    // setCopyThrow
    // Pre: None.
    // Post:
    //     copyThrow_ == shouldThrow
    // Does not throw (No-Throw Guarantee)
    static void setCopyThrow(bool shouldThrow)
    { copyThrow_ = shouldThrow; }

// ***** Counter: Data Members *****
private:

    static int existing_;     // # of existing objects
    static int maxExisting_;  // Max # of existing objects
    static int ctorCount_;    // # of successful (non-throwing) ctor calls
    static int dctorCount_;   // # of dctor calls
    static int assnCount_;    // # of successful (non-throwing) copy = calls
    static bool copyThrow_;   // true if copy operations (ctor, =) throw

};  // End class Counter

// Definition of static data member of class Counter
int Counter::existing_ = 0;
int Counter::maxExisting_ = 0;
int Counter::ctorCount_ = 0;
int Counter::dctorCount_ = 0;
int Counter::assnCount_ = 0;
bool Counter::copyThrow_ = false;


// operator< (Counter)
// Dummy-ish operator<, forming a strict weak order for Counter class
// Returns false (which is legal for a strict weak order; all objects of
//  type Counter are equivalent).
// Pre: None.
// Post:
//     Return value == false.
// Does not throw (No-Throw Guarantee)
bool operator<(const Counter & a,
               const Counter & b)
{ return false; }


// ************************************************************************
// Test Suite Functions
// ************************************************************************


// lex_out
// Print lexeme
void lex_out(const Lex2::Lexeme & it)
{
    std::cout << "        ";
    switch (it.second)
    {
        case Lex2::NONE:
            std::cout << "NONE";
            break;
        case Lex2::ILLEGAL:
            std::cout << "ILLEGAL";
            break;
        case Lex2::KEY:
            std::cout << "KEY";
            break;
        case Lex2::ID:
            std::cout << "ID";
            break;
        case Lex2::OP:
            std::cout << "OP";
            break;
        case Lex2::PUNCT:
            std::cout << "PUNCT";
            break;
        case Lex2::INT:
            std::cout << "INT";
            break;
        case Lex2::FLOAT:
            std::cout << "FLOAT";
            break;
        default:
            std::cout << "<undefined token>";
    }
    std::cout << " [" << it.first << "]" << std::endl;
}


// lex_it
// Test class Lex2 output. Is given initialized object
// and vector of expected lexemes.
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void lex_it(Tester & t,
            Lex2 & lexer,
            const std::vector<Lex2::Lexeme> & correctOutput,
            const std::string & testName)
{
    std::size_t i = 0;
    int err = 0;  // 0: no error, 1: too few, 2: bad lexeme, 3: too many
    std::vector<Lex2::Lexeme> actualOutput;

    while (i < correctOutput.size())
    {
        if (lexer.done())
        {
            err = 1;
            break;
        }
        if (lexer.current().second != Lex2::NONE)
        {
            actualOutput.push_back(lexer.current());
        }
        if (lexer.current() != correctOutput[i])
        {
            err = 2;
            break;
        }

        ++i;
        lexer.advance();
    }

    if (err == 0)
    {
        if (!lexer.done()
          || lexer.current() != Lex2::Lexeme("", Lex2::NONE))
        {
            err = 3;
        }
    }

    t.test(err == 0, testName);
    if (err == 1)
    {
        std::cout << "      Too few lexemes returned" << std::endl;
    }
    else if (err == 3)
    {
        std::cout << "      Too many lexemes returned" << std::endl;
    }
    if (err != 0)
    {
        std::cout << "      Expected lexemes:" << std::endl;
        for (std::size_t k1 = 0; k1 != correctOutput.size(); ++k1)
            lex_out(correctOutput[k1]);
        std::cout << "      Returned lexemes:" << std::endl;
        for (std::size_t k2 = 0; k2 != actualOutput.size(); ++k2)
            lex_out(actualOutput[k2]);
        if (err == 3)
            std::cout << "        (and others)" << std::endl;
    }
#ifdef SINGLE_TEST
    if (err != 0)
        throw std::exception();
#endif //#ifdef SINGLE_TEST
}


// test_class_Lex2_simple
// Test suite for class Lex2, simple inputs
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_simple(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - simple inputs"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    lexer.set("");
    lex_it(t, lexer, output, "empty");

    output.clear();
    lexer.set("  \t\t \r\f\v \n\n");
    lex_it(t, lexer, output, "space only");

    output.clear();
    output.push_back(Lex2::Lexeme("3", Lex2::INT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("b", Lex2::ID));
    lexer.set("3+b");
    lex_it(t, lexer, output, "3+b");
}

// test_class_Lex2_comment
// Test suite for class Lex2, comments
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_comment(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - comments"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("#", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("123", Lex2::INT));
    output.push_back(Lex2::Lexeme("b", Lex2::ID));
    lexer.set("a#123\nb");
    lex_it(t, lexer, output, "# comments not handled");

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("123", Lex2::INT));
    output.push_back(Lex2::Lexeme("b", Lex2::ID));
    lexer.set("a//123\nb");
    lex_it(t, lexer, output, "C++-style comments not handled");

    output.clear();
    lexer.set("/* 123 * / */");
    lex_it(t, lexer, output, "comment only");

    output.clear();
    output.push_back(Lex2::Lexeme("ab", Lex2::ID));
    output.push_back(Lex2::Lexeme("cd", Lex2::ID));
    lexer.set("ab/**/cd");
    lex_it(t, lexer, output, "comment + other");

    output.clear();
    output.push_back(Lex2::Lexeme("ab", Lex2::ID));
    output.push_back(Lex2::Lexeme("ef", Lex2::ID));
    lexer.set("ab/*\ncd\n*/ef");
    lex_it(t, lexer, output, "multi-line comment");

    output.clear();
    output.push_back(Lex2::Lexeme("abc", Lex2::ID));
    lexer.set("abc /* xyz");
    lex_it(t, lexer, output, "non-terminated comment");

    output.clear();
    output.push_back(Lex2::Lexeme("abc", Lex2::ID));
    lexer.set("abc /*/ xyz");
    lex_it(t, lexer, output, "incorrectly terminated comment");

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("b", Lex2::ID));
    output.push_back(Lex2::Lexeme("c", Lex2::ID));
    lexer.set("a /*123*/b/*456*/c");
    lex_it(t, lexer, output, "Two comments");

    output.clear();
    output.push_back(Lex2::Lexeme("x", Lex2::ID));
    output.push_back(Lex2::Lexeme("*", Lex2::OP));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("y", Lex2::ID));
    lexer.set("x*/y");
    lex_it(t, lexer, output, "Comment termination only");
}


// test_class_Lex2_illegal
// Test suite for class Lex2, illegal characters
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_illegal(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - illegal characters"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("\177", Lex2::ILLEGAL));
    lexer.set("\177");
    lex_it(t, lexer, output, "illegal char #1");

    output.clear();
    output.push_back(Lex2::Lexeme("\a", Lex2::ILLEGAL));
    lexer.set("\a");
    lex_it(t, lexer, output, "illegal char #2");

    output.clear();
    output.push_back(Lex2::Lexeme("\a", Lex2::ILLEGAL));
    output.push_back(Lex2::Lexeme("\a", Lex2::ILLEGAL));
    output.push_back(Lex2::Lexeme("\a", Lex2::ILLEGAL));
    output.push_back(Lex2::Lexeme("\a", Lex2::ILLEGAL));
    lexer.set("\a\a/**/\a \a");
    lex_it(t, lexer, output, "multiple illegal chars");

    output.clear();
    output.push_back(Lex2::Lexeme("ab", Lex2::ID));
    output.push_back(Lex2::Lexeme("23", Lex2::INT));
    lexer.set("ab/*\a*/23");
    lex_it(t, lexer, output, "illegal char in comment");

    output.clear();
    output.push_back(Lex2::Lexeme("ab", Lex2::ID));
    output.push_back(Lex2::Lexeme("\a", Lex2::ILLEGAL));
    output.push_back(Lex2::Lexeme("23", Lex2::INT));
    lexer.set("ab\a23");
    lex_it(t, lexer, output, "mixed legal/illegal");
}


// test_class_Lex2_key
// Test suite for class Lex2, keywords
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_key(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - keywords"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    lexer.set("print");
    lex_it(t, lexer, output, "print");

    output.clear();
    output.push_back(Lex2::Lexeme("end", Lex2::KEY));
    lexer.set("end");
    lex_it(t, lexer, output, "end");

    output.clear();
    output.push_back(Lex2::Lexeme("func", Lex2::ID));
    lexer.set("func");
    lex_it(t, lexer, output, "func");

    output.clear();
    output.push_back(Lex2::Lexeme("return", Lex2::ID));
    lexer.set("return");
    lex_it(t, lexer, output, "return");

    output.clear();
    output.push_back(Lex2::Lexeme("PRINT", Lex2::ID));
    lexer.set("PRINT");
    lex_it(t, lexer, output, "PRINT");

    output.clear();
    output.push_back(Lex2::Lexeme("pr", Lex2::ID));
    output.push_back(Lex2::Lexeme("int", Lex2::ID));
    lexer.set("pr int");
    lex_it(t, lexer, output, "pr int");

    output.clear();
    output.push_back(Lex2::Lexeme("pr", Lex2::ID));
    output.push_back(Lex2::Lexeme("int", Lex2::ID));
    lexer.set("pr/**/int");
    lex_it(t, lexer, output, "pr/**/int");

    output.clear();
    output.push_back(Lex2::Lexeme("print2", Lex2::ID));
    lexer.set("print2");
    lex_it(t, lexer, output, "print2");

    output.clear();
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    lexer.set("2print");
    lex_it(t, lexer, output, "2print");

    output.clear();
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    lexer.set("print.");
    lex_it(t, lexer, output, "print.");

    output.clear();
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    lexer.set(".print");
    lex_it(t, lexer, output, ".print");

    output.clear();
    output.push_back(Lex2::Lexeme("printend", Lex2::ID));
    lexer.set("printend");
    lex_it(t, lexer, output, "printend");

    output.clear();
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    output.push_back(Lex2::Lexeme("end", Lex2::KEY));
    lexer.set("print end");
    lex_it(t, lexer, output, "print end");
}


// test_class_Lex2_id
// Test suite for class Lex2, identifiers
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_id(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - identifiers"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    lexer.set("a");
    lex_it(t, lexer, output, "a");

    output.clear();
    output.push_back(Lex2::Lexeme("_a2_a2_a2", Lex2::ID));
    lexer.set("_a2_a2_a2");
    lex_it(t, lexer, output, "_a2_a2_a2");

    output.clear();
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    output.push_back(Lex2::Lexeme("_a2_a2_a", Lex2::ID));
    lexer.set("2_a2_a2_a");
    lex_it(t, lexer, output, "2_a2_a2_a");

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("b", Lex2::ID));
    lexer.set("a b");
    lex_it(t, lexer, output, "a b");

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("b", Lex2::ID));
    lexer.set("a/**/b");
    lex_it(t, lexer, output, "a/**/b");
}


// test_class_Lex2_op
// Test suite for class Lex2, operators
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_op(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - operators"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    lexer.set("+");
    lex_it(t, lexer, output, "+");

    output.clear();
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    lexer.set("-");
    lex_it(t, lexer, output, "-");

    output.clear();
    output.push_back(Lex2::Lexeme("*", Lex2::OP));
    lexer.set("*");
    lex_it(t, lexer, output, "*");

    output.clear();
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    lexer.set("/");
    lex_it(t, lexer, output, "/");

    output.clear();
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("=");
    lex_it(t, lexer, output, "=");

    output.clear();
    output.push_back(Lex2::Lexeme("==", Lex2::OP));
    lexer.set("==");
    lex_it(t, lexer, output, "==");

    output.clear();
    output.push_back(Lex2::Lexeme("!=", Lex2::OP));
    lexer.set("!=");
    lex_it(t, lexer, output, "!=");

    output.clear();
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    lexer.set(".");
    lex_it(t, lexer, output, ".");

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    lexer.set("++");
    lex_it(t, lexer, output, "++");

    output.clear();
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    lexer.set("--");
    lex_it(t, lexer, output, "--");

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("+=");
    lex_it(t, lexer, output, "+=");

    output.clear();
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("-=");
    lex_it(t, lexer, output, "-=");

    output.clear();
    output.push_back(Lex2::Lexeme("==", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("===");
    lex_it(t, lexer, output, "===");

    output.clear();
    output.push_back(Lex2::Lexeme("!=", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("!==");
    lex_it(t, lexer, output, "!==");

    output.clear();
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("= =");
    lex_it(t, lexer, output, "= =");

    output.clear();
    output.push_back(Lex2::Lexeme("(", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("*", Lex2::OP));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("==", Lex2::OP));
    output.push_back(Lex2::Lexeme("!=", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    output.push_back(Lex2::Lexeme(")", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    lexer.set("(+-*/==!==).=");
    lex_it(t, lexer, output, "(+=*/==!+=).=");
}


// test_class_Lex2_punct
// Test suite for class Lex2, punctuation
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_punct(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - punctuation"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    lexer.set(";");
    lex_it(t, lexer, output, ";");

    output.clear();
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("%", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(")", Lex2::PUNCT));
    lexer.set(";;%)");
    lex_it(t, lexer, output, ";;%)");

    output.clear();
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("%", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(")", Lex2::PUNCT));
    lexer.set("; %)");
    lex_it(t, lexer, output, "; %)");
}


// test_class_Lex2_int
// Test suite for class Lex2, integers
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_int(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - integers"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("3", Lex2::INT));
    lexer.set("3");
    lex_it(t, lexer, output, "3");

    output.clear();
    output.push_back(Lex2::Lexeme("00123456789", Lex2::INT));
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    lexer.set("00123456789a");
    lex_it(t, lexer, output, "00123456789a");

    output.clear();
    output.push_back(Lex2::Lexeme("+33", Lex2::INT));
    lexer.set("+33");
    lex_it(t, lexer, output, "+33");

    output.clear();
    output.push_back(Lex2::Lexeme("-33", Lex2::INT));
    lexer.set("-33");
    lex_it(t, lexer, output, "-33");

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("+33", Lex2::INT));
    lexer.set("++33");
    lex_it(t, lexer, output, "++33");

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("-33", Lex2::INT));
    lexer.set("+-33");
    lex_it(t, lexer, output, "+-33");

    output.clear();
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("-33", Lex2::INT));
    lexer.set("--33");
    lex_it(t, lexer, output, "--33");

    output.clear();
    output.push_back(Lex2::Lexeme("33.", Lex2::FLOAT));
    lexer.set("33.");
    lex_it(t, lexer, output, "33.");

    output.clear();
    output.push_back(Lex2::Lexeme("12", Lex2::INT));
    output.push_back(Lex2::Lexeme("34", Lex2::INT));
    lexer.set("12 34");
    lex_it(t, lexer, output, "12 34");
}


// test_class_Lex2_float
// Test suite for class Lex2, floating-point
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_float(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - floating-point"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("1234.56", Lex2::FLOAT));
    lexer.set("1234.56");
    lex_it(t, lexer, output, "1234.56");

    output.clear();
    output.push_back(Lex2::Lexeme(".1234", Lex2::FLOAT));
    lexer.set(".1234");
    lex_it(t, lexer, output, ".1234");

    output.clear();
    output.push_back(Lex2::Lexeme("5678.", Lex2::FLOAT));
    lexer.set("5678.");
    lex_it(t, lexer, output, "5678.");

    output.clear();
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(".5678", Lex2::FLOAT));
    lexer.set("..5678");
    lex_it(t, lexer, output, "..5678");

    output.clear();
    output.push_back(Lex2::Lexeme("5678.", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    lexer.set("5678..");
    lex_it(t, lexer, output, "5678..");

    output.clear();
    output.push_back(Lex2::Lexeme("12.34", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".56", Lex2::FLOAT));
    lexer.set("12.34.56");
    lex_it(t, lexer, output, "12.34.56");

    output.clear();
    output.push_back(Lex2::Lexeme("12.34", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("56.78", Lex2::FLOAT));
    lexer.set("12.34 56.78");
    lex_it(t, lexer, output, "12.34 56.78");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.2e-3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".4", Lex2::FLOAT));
    lexer.set("+1.2e-3.4");
    lex_it(t, lexer, output, "+1.2e-3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("1.2e-3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".4", Lex2::FLOAT));
    lexer.set("+ 1.2e-3.4");
    lex_it(t, lexer, output, "+ 1.2e-3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1", Lex2::INT));
    output.push_back(Lex2::Lexeme(".2e-3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".4", Lex2::FLOAT));
    lexer.set("+1 .2e-3.4");
    lex_it(t, lexer, output, "+1 .2e-3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    output.push_back(Lex2::Lexeme("e", Lex2::ID));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("3.4", Lex2::FLOAT));
    lexer.set("+1. 2e-3.4");
    lex_it(t, lexer, output, "+1. 2e-3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.2", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("e", Lex2::ID));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("3.4", Lex2::FLOAT));
    lexer.set("+1.2 e-3.4");
    lex_it(t, lexer, output, "+1.2 e-3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.2", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("e", Lex2::ID));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("3.4", Lex2::FLOAT));
    lexer.set("+1.2e -3.4");
    lex_it(t, lexer, output, "+1.2e -3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.2", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("e", Lex2::ID));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("3.4", Lex2::FLOAT));
    lexer.set("+1.2e- 3.4");
    lex_it(t, lexer, output, "+1.2e- 3.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.2e-3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".4", Lex2::FLOAT));
    lexer.set("+1.2e-3 .4");
    lex_it(t, lexer, output, "+1.2e-3 .4");

    output.clear();
    output.push_back(Lex2::Lexeme("+1.2e-3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("4", Lex2::INT));
    lexer.set("+1.2e-3. 4");
    lex_it(t, lexer, output, "+1.2e-3. 4");

    output.clear();
    output.push_back(Lex2::Lexeme("$", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("1.E2", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    lexer.set("$1.E2a");
    lex_it(t, lexer, output, "$1.E2a");

    output.clear();
    output.push_back(Lex2::Lexeme("_1", Lex2::ID));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("E2a", Lex2::ID));
    lexer.set("_1.E2a");
    lex_it(t, lexer, output, "_1.E2a");

    output.clear();
    output.push_back(Lex2::Lexeme("$", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("1.", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("EE2a", Lex2::ID));
    lexer.set("$1.EE2a");
    lex_it(t, lexer, output, "$1.EE2a");

    output.clear();
    output.push_back(Lex2::Lexeme("$", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("1.", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("E", Lex2::ID));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    lexer.set("$1.E--a");
    lex_it(t, lexer, output, "$1.E--2a");

    output.clear();
    output.push_back(Lex2::Lexeme("+.3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("x", Lex2::ID));
    output.push_back(Lex2::Lexeme("$", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("x", Lex2::ID));
    lexer.set("+.3 +.x$+.x");
    lex_it(t, lexer, output, "+.3 +.x$+.x");

    output.clear();
    output.push_back(Lex2::Lexeme("1.", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("e", Lex2::ID));
    output.push_back(Lex2::Lexeme(".4", Lex2::FLOAT));
    lexer.set("1.e.4");
    lex_it(t, lexer, output, "1.e.4");

    output.clear();
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme(".", Lex2::PUNCT));
    lexer.set("+.");
    lex_it(t, lexer, output, "+.");

    output.clear();
    output.push_back(Lex2::Lexeme("1.2e3", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("1.2", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("f3", Lex2::ID));
    lexer.set("1.2e3 1.2f3");
    lex_it(t, lexer, output, "1.2e3 1.2f3");
}


// test_class_Lex2_plus_minus
// Test suite for class Lex2, plus/minus rule
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_plus_minus(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - plus/minus rule"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    output.push_back(Lex2::Lexeme("+2", Lex2::INT));
    lexer.set("print+2");
    lex_it(t, lexer, output, "print+2");

    output.clear();
    output.push_back(Lex2::Lexeme("abc", Lex2::ID));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    lexer.set("abc+2");
    lex_it(t, lexer, output, "abc+2");

    output.clear();
    output.push_back(Lex2::Lexeme("==", Lex2::OP));
    output.push_back(Lex2::Lexeme("+2", Lex2::INT));
    lexer.set("==+2");
    lex_it(t, lexer, output, "==+2");

    output.clear();
    output.push_back(Lex2::Lexeme("%", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("+2", Lex2::INT));
    lexer.set("%+2");
    lex_it(t, lexer, output, "%+2");

    output.clear();
    output.push_back(Lex2::Lexeme("22", Lex2::INT));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    lexer.set("22-2");
    lex_it(t, lexer, output, "22-2");

    output.clear();
    output.push_back(Lex2::Lexeme("2.2", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    lexer.set("2.2+2");
    lex_it(t, lexer, output, "2.2+2");

    output.clear();
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    output.push_back(Lex2::Lexeme("$", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("+2", Lex2::INT));
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("2", Lex2::INT));
    output.push_back(Lex2::Lexeme("$", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("+2", Lex2::INT));
    lexer.set("a+2 $+2 a+2 $+2");
    lex_it(t, lexer, output, "a+2 $+2 a+2 $+2");
}


// test_class_Lex2_long
// Test suite for class Lex2, long tests
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2_long(Tester & t)
{
    std::cout << "Test Suite: class Lex2 - long tests"
              << std::endl;

    Lex2 lexer;
    std::vector<Lex2::Lexeme> output;

    output.clear();
    output.push_back(Lex2::Lexeme("for", Lex2::ID));
    output.push_back(Lex2::Lexeme("(", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("int", Lex2::ID));
    output.push_back(Lex2::Lexeme("i", Lex2::ID));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    output.push_back(Lex2::Lexeme("0", Lex2::INT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("i", Lex2::ID));
    output.push_back(Lex2::Lexeme("!=", Lex2::OP));
    output.push_back(Lex2::Lexeme("43", Lex2::INT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("i", Lex2::ID));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    output.push_back(Lex2::Lexeme("3", Lex2::INT));
    output.push_back(Lex2::Lexeme(")", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("{", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("string", Lex2::ID));
    output.push_back(Lex2::Lexeme("s", Lex2::ID));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    output.push_back(Lex2::Lexeme("\"", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("abc", Lex2::ID));
    output.push_back(Lex2::Lexeme("\"", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("float", Lex2::ID));
    output.push_back(Lex2::Lexeme("f", Lex2::ID));
    output.push_back(Lex2::Lexeme("=", Lex2::OP));
    output.push_back(Lex2::Lexeme("3.2e43", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    output.push_back(Lex2::Lexeme("something", Lex2::ID));
    output.push_back(Lex2::Lexeme("cout", Lex2::ID));
    output.push_back(Lex2::Lexeme("<", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("<", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("s", Lex2::ID));
    output.push_back(Lex2::Lexeme("<", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("<", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("f", Lex2::ID));
    output.push_back(Lex2::Lexeme("<", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("<", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("endl", Lex2::ID));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("}", Lex2::PUNCT));
    lexer.set(
        "for (int i=0; i!=43; i+=3)\n"
        "{\n"
        "    /* The body of the loop */\n"
        "    string s = \"abc\";\n"
        "    float f=3.2e43;\n"
        "    // print something\n"
        "    cout << s << f << endl;\n"
        "}\n"
        );
    lex_it(t, lexer, output, "C++ code snippet");

    output.clear();
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    output.push_back(Lex2::Lexeme("1", Lex2::INT));
    output.push_back(Lex2::Lexeme("*", Lex2::OP));
    output.push_back(Lex2::Lexeme("(", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("2.4", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("5", Lex2::INT));
    output.push_back(Lex2::Lexeme("-", Lex2::OP));
    output.push_back(Lex2::Lexeme("a", Lex2::ID));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("3", Lex2::INT));
    output.push_back(Lex2::Lexeme("/", Lex2::OP));
    output.push_back(Lex2::Lexeme("2.1e7", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme(")", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("print", Lex2::KEY));
    output.push_back(Lex2::Lexeme("(", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("34.5", Lex2::FLOAT));
    output.push_back(Lex2::Lexeme("+", Lex2::OP));
    output.push_back(Lex2::Lexeme("-58", Lex2::INT));
    output.push_back(Lex2::Lexeme(")", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme(";", Lex2::PUNCT));
    output.push_back(Lex2::Lexeme("end", Lex2::KEY));
    lexer.set(
        "/* A little program */\n"
        "print 1*(2.4+5-a+3/2.1e7);\n"
        "print (34.5+-58);\n"
        "end\n"
        );
    lex_it(t, lexer, output, "other code snippet");

    std::string s;
    bool success;
    
    s = "";
    output.clear();
    for (int i = 0; i < 100000; ++i)
        s += "XYZ";
    output.push_back(Lex2::Lexeme(s, Lex2::ID));
    lexer.set(s);
    lex_it(t, lexer, output, "long identifier");

    s = "";
    output.clear();
    for (int i = 0; i < 100000; ++i)
        s += "123";
    output.push_back(Lex2::Lexeme(s, Lex2::INT));
    lexer.set(s);
    lex_it(t, lexer, output, "long integer");

    // Long sequence of punctuation
    // Expect PUNCT, PUNCT, PUNCT, ... , PUNCT, NONE, NONE
    s = "";
    for (int i = 0; i < 100000; ++i)
        s += "$";
    lexer.set(s);
    success = true;
    for (int i = 0; i < 100000; ++i)
    {
        if (lexer.current() != Lex2::Lexeme("$", Lex2::PUNCT))
        {
            success = false;
            break;
        }
        lexer.advance();
    }
    if (success)
    {
        if (lexer.current() != Lex2::Lexeme("", Lex2::NONE))
        {
            success = false;
        }
    }
    if (success)
    {
        lexer.advance();
        if (lexer.current() != Lex2::Lexeme("", Lex2::NONE))
        {
            success = false;
        }
    }
    t.test(success, "Long sequence of punctuation");

    // Read past end
    // Expect INT, NONE, NONE, NONE, NONE, NONE, ...
    s = "1";
    lexer.set(s);
    success = true;
    if (lexer.current() != Lex2::Lexeme("1", Lex2::INT))
    {
    s = "";
    output.clear();
    for (int i = 0; i < 100000; ++i)
        s += "123";
    output.push_back(Lex2::Lexeme(s, Lex2::INT));
    lexer.set(s);
    lex_it(t, lexer, output, "long integer");

        t.test(false, "Read far past end");
        success = false;
    }
    else
    {
        for (int i = 0; i < 100000; ++i)
        {
            lexer.advance();
            if (lexer.current() != Lex2::Lexeme("", Lex2::NONE))
            {
                success = false;
                break;
            }
        }
    }
    t.test(success, "Read far past end");
}


// test_class_Lex2
// Test suite for class Lex2
// Uses other test-suite functions
// Pre: None.
// Post:
//     Pass/fail status of tests have been registered with t.
//     Appropriate messages have been printed to cout.
// If SINGLE_TEST not defined:
//     Does not throw (No-Throw Guarantee)
// If SINGLE_TEST defined:
//     Throws std::exception on test failure; Basic Guarantee
void test_class_Lex2(Tester & t)
{
    // Do all the test suites
    std::cout << "TEST SUITES FOR CLASS Lex2" << std::endl;
    test_class_Lex2_simple(t);
    test_class_Lex2_comment(t);
    test_class_Lex2_illegal(t);
    test_class_Lex2_key(t);
    test_class_Lex2_id(t);
    test_class_Lex2_op(t);
    test_class_Lex2_punct(t);
    test_class_Lex2_int(t);
    test_class_Lex2_float(t);
    test_class_Lex2_plus_minus(t);
    test_class_Lex2_long(t);
}


// ************************************************************************
// Main program
// ************************************************************************


// main
// Runs class Lex2 test suite, prints results to cout.
int main()
{
    Tester t;
#ifdef SINGLE_TEST
    try
    {
        test_class_Lex2(t);
    }
    catch (std::exception & e)
    { /* Nothing to do here ... */ }
#else //#ifdef SINGLE_TEST
    test_class_Lex2(t);
#endif //#ifdef SINGLE_TEST

    std::cout << std::endl;
    if (t.allPassed())
    {
        std::cout << "All tests successful" 
                  << std::endl;
    }
    else
    {
        std::cout << "Tests ********** UNSUCCESSFUL **********"
                  << std::endl;
#ifdef SINGLE_TEST
        std::cout << std::endl;
        std::cout << "NOTE: THIS PROGRAM TERMINATES AFTER THE FIRST"
                  << std::endl;
        std::cout << "UNSUCCESSFUL TEST (IF ANY). TO CHANGE THIS"
                  << std::endl;
        std::cout << "BEHAVIOR, COMMENT OUT THE LINE DEFINING THE"
                  << std::endl;
        std::cout << "CONSTANT \"SINGLE_TEST\"."
                  << std::endl;
#endif //#ifdef SINGLE_TEST
    }
    std::cout << std::endl;

    std::cout << "Press ENTER to quit ";
    while (std::cin.get() != '\n') ;

    return 0;
}

